//***********************************************************************************
//
//    Bendy Tentacle Demo
//    (c) 2001 NCCA
//    by Eike F Anderson   ( mailto:eike@liquid5.co.uk )
//
//    Real-Time Skeletal Animation, Mesh Skinning and IK
//
//    Revision# 12
//    Last Changed: 30/03/2001
//
//    API: OpenGL
//
//    this file contains sourcecode written by Jeff Molofee
//
//    Permission to use, copy, modify and distribute for non-commercial purposes
//    is granted without fee, as long as both, the above copyright notice and this
//    permission notice appear in all copies/modifications and supporting
//    documentation.
//    Modified versions of this code MUST be clearly marked as such.
//    
//    This software is provided "as is" without any expressed or implied guarantee
//    or warranty.
//
//    this software is Freeware (NOT Public Domain)
//
//***********************************************************************************

#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <GL\glaux.h>	
#include <math.h>
#include <stdio.h>

#define  COS(A)   (cosf((A)*0.017453292519943295769236907684886))
#define  SIN(A)   (sinf((A)*0.017453292519943295769236907684886))
#define  NUM_VERTICES  49
#define  NUM_BONES      2

AUX_RGBImageRec *LoadBMP(char *);		  // Loads A Bitmap Image
int LoadGLTextures();					  // Load Bitmaps And Convert To Textures
void matxmat(GLfloat*,GLfloat*,GLfloat*); // Matrix-Matrix Multiplication
void matxvec(GLfloat*,GLfloat*,GLfloat*); // Vector-Matrix Multiplication
int calculate_normals(GLfloat[][3],int,GLint[][3],int,GLfloat[][3]);  // generate the mesh normals
void setup();                             // initialise data
void build_matrices();                    // generate the local transformation matrices
void skin();                              // mesh skinning function - WorkInProgress
int WINAPI WinMain(HINSTANCE,HINSTANCE,LPSTR,int); // program entry function
void SetDCPixelFormat(HWND,HDC);          // prepare Window for OpenGL drawing
void InitializeRC(void);                  // setup the scene
void makeMaterial(GLenum,GLfloat*,GLfloat);  // set the current material
void setMaterial(int);                    // select a material
void DrawScene(HDC,double);               // draw the current frame
double cosEnclosed(GLfloat*,GLfloat*);    // cosine of angle between 2 vectors
double sinEnclosed(GLfloat*,GLfloat*);    // sine of angle between 2 vectors
double magnitude(GLfloat*);               // magnitude (length) of vector
int sign(double);                         // sign of value
double solvejoint(GLfloat*,GLfloat*,GLfloat*); // IK solver
void chain_up(int,GLfloat*,GLfloat*);     // calculate extent of IK chain

struct materials{
 GLfloat ambient[3];      // ambient colour of material
 GLfloat diffuse[3];      // diffuse colour of material
 GLfloat specular[3];     // specular colour of material
 GLfloat glow[3];         // light emitted from material
 GLfloat alpha;           // alpha value
 GLfloat shine;           // specular exponent (shinyness)
 // perhaps include a texture field?!
};


struct bone{
  double rotx;           // local rotation values
  double roty;
  double rotz;
  GLfloat xoffset;       // local offset (in relation to parent)
  GLfloat yoffset;
  GLfloat zoffset;
  GLfloat xextens;       // extension of bone (needed for IK)
  GLfloat yextens;
  GLfloat zextens;
  GLfloat xjointx;       // joint X rotation axis
  GLfloat xjointy;
  GLfloat xjointz;
  GLfloat zjointx;       // joint Z rotation axis
  GLfloat zjointy;
  GLfloat zjointz;
  GLfloat minX;			 // limits - not yet implemented
  GLfloat maxX;
  GLfloat minY;
  GLfloat maxY;
  GLfloat minZ;
  GLfloat maxZ;
  // one could add an IK weight factor for each bone - like a spring...
  int mat;               // bone material
  GLfloat tmatrix[16];   // local transformation matrix
  int parent;            // index of parent bone
};

struct vertex
{
	GLfloat x;        // skin vertex x coordinate
	GLfloat y;        // skin vertex y coordinate
	GLfloat z;        // skin vertex z coordinate
	GLfloat nx;       // skin vertex normal x coordinate
	GLfloat ny;       // skin vertex normal y coordinate
	GLfloat nz;       // skin vertex normal z coordinate
	GLfloat u;        // skin vertex texture coordinate u
	GLfloat v;        // skin vertex texture coordinate v
	int bone1;        // first influential bone
	int bone2;        // second influential bone
	GLfloat weight;   // weighting percentage of bone1
	                  // percentage for bone2 can be calculated:
	                  // bone2_weight = 1-weight
};

bone bones[NUM_BONES];// bones

GLuint	texture[2];   // texture
GLfloat IKtarget[3];  // IK target coordinates
int window_width=800,window_height=600;


//  flags
int bone_visibility=1,turntable=1,action=1,optperf=0,limits=0;
int normal_visibility=0, bones_off=0,fwkin=0,invkin=0;
// bone_visibility   - wireframe mode or shaded mode
// turntable         - rotate scene
// action            - animate
// optperf           - use textures
// limits            - use constraints for bone motion - not yet implemented
// normal_visibility - display normals
// bones_off         - display of bones in wireframe mode?
// fwkin             - use Forward Kinematics
// invkin            - use Inverse Kinematics

// speed settings
double speed1=1.0f,speed2=0.1f;

static materials materials [7] = {
// floor
// bone0
// bone1
// tentacle (shaded)
// normals
// tentacle (wireframe)
// IK pull vector
 {{0.8f,0.8f,0.8f},	{0.752941f,0.752941f,0.752941f},	{0.9f,0.9f,0.9f},	{0.3f,0.3f,0.3f},	1.0f,8.0f},
 {{0.0f,0.0f,1.0f},	{0.0,0.0f,1.0f},	{0.0,0.0f,1.0f},	{0.0f,0.0f,1.0f},	1.0f,8.0f},
 {{0.0f,0.3f,1.0f},	{0.0,0.3f,1.0f},	{0.0,0.3f,1.0f},	{0.0f,0.3f,1.0f},	1.0f,8.0f},
 {{0.9f,1.0f,0.9f},	{0.8f,1.0f,0.8f},	{0.8f,1.0f,0.8f},	{0.4f,0.7f,0.4f},	1.0f,8.0f},
 {{0.8f,1.0f,0.2f},	{0.8f,1.0f,0.2f},	{0.8f,1.0f,0.2f},	{0.8f,0.5f,0.0f},	1.0f,8.0f},
 {{0.0f,1.0f,0.117647f},	{0.0f,1.0f,0.752941f},	{0.0f,1.0f,0.752941f},	{0.0f,0.5f,0.0f},	1.0f,8.0f},
 {{1.0f,0.0f,0.0f},	{1.0,0.0f,0.0f},	{1.0,0.0f,0.0f},	{1.0f,0.0f,0.0f},	1.0f,8.0f}
};

static GLint floor_indices[2][9] = {
// floor vertex indices
	{0,1,2 ,0,1,2 ,0,1,2 }, {1,3,2 ,3,4,5 ,3,4,5 }
};
static GLfloat floor_vertices [4][3] = {
{-2.5f,0.0f,2.5f},{2.5f,0.0f,2.5f},{-2.5f,0.0f,-2.5f},
{2.5f,0.0f,-2.5f}
};
static GLfloat floor_normals [6][3] = {
{0.0f,1.0f,0.0f},{0.0f,1.0f,0.0f},{0.0f,1.0f,0.0f},
{0.0f,1.0f,0.0f},{0.0f,1.0f,0.0f},{0.0f,1.0f,0.0f}
};
static GLfloat floor_texture [6][2] = {
{0.0f,0.0f},{1.0f,0.0f},{0.0f,1.0f},
{1.0f,0.0f},{1.0f,1.0f},{0.0f,1.0f}
};


static GLint bone_indices[4][9] = {
// bone vertex indices
	{0,2,1 ,0,1,2 ,0,1,2 }, {0,1,3 ,3,4,5 ,3,4,5 }, {1,2,3 ,6,7,8 ,6,7,8 },
	{2,0,3 ,9,10,11 ,9,10,11 }
};
static GLfloat bone_vertices [4][3] = {
{-0.05f,0.0f,-0.086603f},{-0.05f,0.0f,0.086603f},{0.1f,0.0f,0.0f},
{0.0f,1.5f,0.0f}
};
static GLfloat bone_normals [12][3] = {
{0.0f,-1.0f,0.0f},{0.0f,-1.0f,0.0f},{0.0f,-1.0f,0.0f},
{-0.999445f,0.0333148f,0.0f},{-0.999445f,0.0333148f,0.0f},{-0.999445f,0.0333148f,0.0f},
{0.499724f,0.033315f,0.865544f},{0.499724f,0.033315f,0.865544f},{0.499724f,0.033315f,0.865544f},
{0.499724f,0.033315f,-0.865544f},{0.499724f,0.033315f,-0.865544f},{0.499724f,0.033315f,-0.865544f}
};
static GLfloat bone_textures [12][2] = {
{0.25f,0.066987f},{1.0f,0.5f},{0.25f,0.933013f},
{0.0f,0.0f},{0.333333f,0.0f},{0.5f,1.0f},
{0.333333f,0.0f},{0.666667f,0.0f},{0.5f,1.0f},
{0.666667f,0.0f},{1.0f,0.0f},{0.5f,1.0f}
};

GLfloat normals[NUM_VERTICES][3];               // normals
GLint   tentacle_indices[88][3];				// indices

vertex  tentacle_skin[NUM_VERTICES];            // skin vertices

static GLfloat vertices[NUM_VERTICES][3]=
// the 49 vertices that make up the tentacle
{
 {  0.0f,  3.0f,  0.0f},
 {  0.1f,  2.9f,  0.0f},
 { 0.07f,  2.9f, 0.07f},
 {  0.0f,  2.9f,  0.1f},
 {-0.07f,  2.9f, 0.07f},
 { -0.1f,  2.9f,  0.0f},
 {-0.07f,  2.9f,-0.07f},
 {  0.0f,  2.9f, -0.1f},
 { 0.07f,  2.9f,-0.07f},
 { 0.24f,  2.0f,  0.0f},
 { 0.18f,  2.0f, 0.18f},
 {  0.0f,  2.0f, 0.24f},
 {-0.18f,  2.0f, 0.18f},
 {-0.24f,  2.0f,  0.0f},
 {-0.18f,  2.0f,-0.18f},
 {  0.0f,  2.0f,-0.24f},
 { 0.18f,  2.0f,-0.18f},
 { 0.33f,  1.5f,  0.0f},
 { 0.23f,  1.5f, 0.23f},
 {  0.0f,  1.5f, 0.33f},
 {-0.23f,  1.5f, 0.23f},
 {-0.33f,  1.5f,  0.0f},
 {-0.23f,  1.5f,-0.23f},
 {  0.0f,  1.5f,-0.33f},
 { 0.23f,  1.5f,-0.23f},
 { 0.44f,  1.0f,  0.0f},
 { 0.31f,  1.0f, 0.31f},
 {  0.0f,  1.0f, 0.44f},
 {-0.31f,  1.0f, 0.31f},
 {-0.44f,  1.0f,  0.0f},
 {-0.31f,  1.0f,-0.31f},
 {  0.0f,  1.0f,-0.44f},
 { 0.31f,  1.0f,-0.31f},
 { 0.62f,  0.2f,  0.0f},  // the normals for these seem to be fine
 { 0.44f,  0.2f, 0.44f},
 {  0.0f,  0.2f, 0.62f},
 {-0.44f,  0.2f, 0.44f},
 {-0.62f,  0.2f,  0.0f},
 {-0.44f,  0.2f,-0.44f},
 {  0.0f,  0.2f,-0.62f},
 { 0.44f,  0.2f,-0.44f},
 { 0.75f,  0.0f,  0.0f},  // bottom level
 { 0.53f,  0.0f, 0.53f},
 {  0.0f,  0.0f, 0.75f},
 {-0.53f,  0.0f, 0.53f},
 {-0.75f,  0.0f,  0.0f},
 {-0.53f,  0.0f,-0.53f},
 {  0.0f,  0.0f,-0.75f},
 { 0.53f,  0.0f,-0.53f}
};

static GLint indices[88][3]=
// the triangle indices for the tentacle
// some may not be quite in the correct order...
// that happens if you calculate them by hand...
{
 {  0,  1,  2},
 {  0,  2,  3},
 {  0,  3,  4},
 {  0,  4,  5},
 {  0,  5,  6},
 {  0,  6,  7},
 {  0,  7,  8},
 {  0,  8,  1},
 
 {  1,  9, 10},
 {  1, 10,  2},
 {  2, 10, 11},
 {  2, 11,  3},
 {  3, 11, 12},
 {  3, 12,  4},
 {  4, 12, 13},
 {  4, 13,  5},
 {  5, 13, 14},
 {  5, 14,  6},
 {  6, 14, 15},
 {  6, 15,  7},
 {  7, 15, 16},
 {  7, 16,  8},
 {  8, 16,  9},
 {  8,  9,  1},
 
 {  9, 17, 18},
 {  9, 18, 10},
 { 10, 18, 19},
 { 10, 19, 11},
 { 11, 19, 20},
 { 11, 20, 12},
 { 12, 20, 21},
 { 12, 21, 13},
 { 13, 21, 22},
 { 13, 22, 14},
 { 14, 22, 23},
 { 14, 23, 15},
 { 15, 23, 24},
 { 15, 24, 16},
 { 16, 24, 17},
 { 16, 17,  9},

 { 17, 25, 26},
 { 17, 26, 18},
 { 18, 26, 27},
 { 18, 27, 19},
 { 19, 27, 28},
 { 19, 28, 20},
 { 20, 28, 29},
 { 20, 29, 21},
 { 21, 29, 30},
 { 21, 30, 22},
 { 22, 30, 31},
 { 22, 31, 23},
 { 23, 31, 32},
 { 23, 32, 24},
 { 24, 32, 25},
 { 24, 25, 17},

 { 25, 33, 34},
 { 25, 34, 26},
 { 26, 34, 35},
 { 26, 35, 27},
 { 27, 35, 36},
 { 27, 36, 28},
 { 28, 36, 37},
 { 28, 37, 29},
 { 29, 37, 38},
 { 29, 38, 30},
 { 30, 38, 39},
 { 30, 39, 31},
 { 31, 39, 40},
 { 31, 40, 32},
 { 32, 40, 33},
 { 32, 33, 25},

 { 33, 41, 42},
 { 33, 42, 34},
 { 34, 42, 43},
 { 34, 43, 35},
 { 35, 43, 44},
 { 35, 44, 36},
 { 36, 44, 45},
 { 36, 45, 37},
 { 37, 45, 46},
 { 37, 46, 38},
 { 38, 46, 47},
 { 38, 47, 39},
 { 39, 47, 48},
 { 39, 48, 40},
 { 40, 48, 41},
 { 40, 41, 33}
};



// the following function was taken from Jeff Molofee's (NeHe-)Tutorial 6
AUX_RGBImageRec *LoadBMP(char *Filename)				// Loads A Bitmap Image
{
	FILE *File=NULL;									// File Handle

	if (!Filename)										// Make Sure A Filename Was Given
	{
		return NULL;									// If Not Return NULL
	}

	File=fopen(Filename,"r");							// Check To See If The File Exists

	if (File)											// Does The File Exist?
	{
		fclose(File);									// Close The Handle
		return auxDIBImageLoad(Filename);				// Load The Bitmap And Return A Pointer
	}

	return NULL;										// If Load Failed Return NULL
}

// the following function was taken from Jeff Molofee's (NeHe-)Tutorial 6
int LoadGLTextures()									// Load Bitmaps And Convert To Textures
{
	int Status=FALSE;									// Status Indicator

	AUX_RGBImageRec *TextureImage[2];					// Create Storage Space For The Texture

	memset(TextureImage,0,sizeof(void *)*2);           	// Set The Pointer To NULL

	// Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit
	if (TextureImage[0]=LoadBMP("tentacle.bmp"))
	{
		Status=TRUE;									// Set The Status To TRUE

		glGenTextures(1, &texture[0]);					// Create The Texture

		// Typical Texture Generation Using Data From The Bitmap
		glBindTexture(GL_TEXTURE_2D, texture[0]);
		glTexImage2D(GL_TEXTURE_2D, 0, 3, TextureImage[0]->sizeX, TextureImage[0]->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage[0]->data);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	}

	if (TextureImage[0])									// If Texture Exists
	{
		if (TextureImage[0]->data)							// If Texture Image Exists
		{
			free(TextureImage[0]->data);					// Free The Texture Image Memory
		}

		free(TextureImage[0]);								// Free The Image Structure
	}
	if (TextureImage[1]=LoadBMP("floor.bmp"))
	{
		Status=TRUE;									// Set The Status To TRUE

		glGenTextures(1, &texture[1]);					// Create The Texture

		// Typical Texture Generation Using Data From The Bitmap
		glBindTexture(GL_TEXTURE_2D, texture[1]);
		glTexImage2D(GL_TEXTURE_2D, 0, 3, TextureImage[1]->sizeX, TextureImage[1]->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage[1]->data);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	}

	if (TextureImage[1])									// If Texture Exists
	{
		if (TextureImage[1]->data)							// If Texture Image Exists
		{
			free(TextureImage[1]->data);					// Free The Texture Image Memory
		}

		free(TextureImage[1]);								// Free The Image Structure
	}

	return Status;										// Return The Status
}

double solvejoint(GLfloat *a,GLfloat *b,GLfloat *f)
{
	GLfloat retval,rlength;
	GLfloat r[3];
	GLfloat modify=0.25f;
    // modify is an interesting value - it could be used as a bone weight for the IK:
	// at 0.25 the end effector slowly follows the IK target
    // the larger this value gets, the faster the end effector moves
	// from about value 3 onwards, the end effector becomes more and more likely to "overshoot"
	// the IK target and to "break" the tentacle

	// calculate a vector that is perpendicular to
	// a (rotation axis) and b (axis of bone) by
	// calculating the cross-product axb
	r[0]=(a[1]*b[2])-(a[2]*b[1]);
	r[1]=(a[2]*b[0])-(a[0]*b[2]);
	r[2]=(a[0]*b[1])-(a[1]*b[0]);
#pragma warning( disable : 4244 )
	// and then normalise the new vector
	rlength=sqrt((r[0]*r[0])+(r[1]*r[1])+(r[2]*r[2]));
	r[0]/=rlength;
	r[1]/=rlength;
	r[2]/=rlength;

    // the formula that solves the problem...
	retval=magnitude(f)*sinEnclosed(a,f)*sinEnclosed(b,f)*sign(cosEnclosed(r,f))*modify;
#pragma warning( default : 4244 )
	return retval;
}

double magnitude(GLfloat *v)
{
#pragma warning( disable : 4244 )
	return sqrt((v[0]*v[0])+(v[1]*v[1])+(v[2]*v[2]));
#pragma warning( default : 4244 )
}

double cosEnclosed(GLfloat *a,GLfloat *b)
{
	GLfloat alength,blength,d;
	d=(a[0]*b[0])+(a[1]*b[1])+(a[2]*b[2]);
#pragma warning( disable : 4244 )
	alength=sqrt((a[0]*a[0])+(a[1]*a[1])+(a[2]*a[2]));
	blength=sqrt((b[0]*b[0])+(b[1]*b[1])+(b[2]*b[2]));
#pragma warning( default : 4244 )
	return d/(alength*blength);
}

double sinEnclosed(GLfloat *a,GLfloat *b)
{
	GLfloat clength, alength, blength;
	GLfloat c[3];
	c[0]=(a[1]*b[2])-(a[2]*b[1]);
	c[1]=(a[2]*b[0])-(a[0]*b[2]);
	c[2]=(a[0]*b[1])-(a[1]*b[0]);
#pragma warning( disable : 4244 )
	alength=sqrt((a[0]*a[0])+(a[1]*a[1])+(a[2]*a[2]));
	blength=sqrt((b[0]*b[0])+(b[1]*b[1])+(b[2]*b[2]));
	clength=sqrt((c[0]*c[0])+(c[1]*c[1])+(c[2]*c[2]));
#pragma warning( default : 4244 )
	return clength/(alength*blength);
}

int sign(double n)
{
	if(n>=0) return 1;
	return (-1);
}

void matxmat(GLfloat *ret,GLfloat *m1,GLfloat *m2)
{
  int i,j;
  for(i=0;i<4;i++)
	for(j=0;j<4;j++)
	{
	  ret[(4*i)+j]=(m1[j]*m2[4*i])+(m1[j+4]*m2[(4*i)+1])+(m1[j+8]*m2[(4*i)+2])+(m1[j+12]*m2[(4*i)+3]);
	}
}

void matxvec(GLfloat *ret,GLfloat *m,GLfloat*v)
{
  int i;
  for(i=0;i<4;i++)
  {
	ret[i]=(m[i]*v[0])+(m[i+4]*v[1])+(m[i+8]*v[2])+(m[i+12]*v[3]);
  }
}

int calculate_normals(GLfloat verts[][3],int vertices,GLint inds[][3],int indices,GLfloat norms[][3])
{
   GLfloat **trinorms;
   int i,j;
   GLfloat vx,vy,vz,wx,wy,wz;
   GLfloat cx,cy,cz,divider;

   trinorms=(GLfloat**)malloc(indices*sizeof(GLfloat*));
   if(trinorms==NULL)
	   return 0;
   for(i=0;i<indices;i++)
   {
     trinorms[i]=(GLfloat*)malloc(3*sizeof(GLfloat));
	 if(trinorms[i]==NULL)
		 return 0;
   }

   for(i=0;i<indices;i++)
   {
     vx=verts[inds[i][0]][0]-verts[inds[i][1]][0];
	 vy=verts[inds[i][0]][1]-verts[inds[i][1]][1];
	 vz=verts[inds[i][0]][2]-verts[inds[i][1]][2];
	 wx=verts[inds[i][1]][0]-verts[inds[i][2]][0];
	 wy=verts[inds[i][1]][1]-verts[inds[i][2]][1];
	 wz=verts[inds[i][1]][2]-verts[inds[i][2]][2];

	 cx=(vy*wz)-(vz*wy);
	 cy=(vz*wx)-(vx*wz);
	 cz=(vx*wy)-(vy*wx);
#pragma warning( disable : 4244 )
     divider=sqrt((cx*cx)+(cy*cy)+(cz*cz));
#pragma warning( default : 4244 )
	 trinorms[i][0]=cx/divider;
	 trinorms[i][1]=cy/divider;
	 trinorms[i][2]=cz/divider;
   }
   int count;
   for(i=0;i<vertices;i++)
   {
	   count=0;
	   cx=0;
	   cy=0;
	   cz=0;
	   for(j=0;j<indices;j++)
	   {
		   if(inds[j][0]==i)
		   {
			   cx+=trinorms[inds[j][0]][0];
			   cy+=trinorms[inds[j][0]][1];
			   cz+=trinorms[inds[j][0]][2];
			   count++;
		   }
		   else if(inds[j][1]==i)
		   {
			   cx+=trinorms[inds[j][1]][0];
			   cy+=trinorms[inds[j][1]][1];
			   cz+=trinorms[inds[j][1]][2];
			   count++;
		   }
		   else if(inds[j][2]==i)
		   {
			   cx+=trinorms[inds[j][2]][0];
			   cy+=trinorms[inds[j][2]][1];
			   cz+=trinorms[inds[j][2]][2];
			   count++;
		   }
	   }
	   cx/=count;
	   cy/=count;
	   cz/=count;
#pragma warning( disable : 4244 )
       divider=sqrt((cx*cx)+(cy*cy)+(cz*cz));
#pragma warning( default : 4244 )
	   norms[i][0]=cx/divider;
	   norms[i][1]=cy/divider;
	   norms[i][2]=cz/divider;
   }

   for(i=0;i<indices;i++)           // free the temporary storage
     free(trinorms[i]);
   free(trinorms);

   return 1;
}

void setup()
//  setup the skeleton and mesh
//  ---------------------------
//  if this wasn't hard-coded this would have to be
//  performed by a file format loader/interpreter
{
  int i;
  // 
  if (!LoadGLTextures())    // load textures
  {
	  optperf=1;            // don't use them
  }

  // prepare skin face indices
  for(i=0;i<88;i++)
  {
    tentacle_indices[i][0]=indices[i][0];
    tentacle_indices[i][1]=indices[i][1];
    tentacle_indices[i][2]=indices[i][2];
  }

  // setup skin
  for(i=0;i<NUM_VERTICES;i++)
  {
	tentacle_skin[i].x=vertices[i][0];
	tentacle_skin[i].y=vertices[i][1];
	tentacle_skin[i].z=vertices[i][2];
	tentacle_skin[i].u=(vertices[i][0]+1)/2;
	tentacle_skin[i].v=(vertices[i][2]+1)/2;
  }
  for(i=0;i<9;i++)
  {
	tentacle_skin[i].weight=0.0f;
  }
  for(i=9;i<17;i++)
  {
	tentacle_skin[i].weight=0.3f;
  }
  for(i=17;i<25;i++)
  {
	tentacle_skin[i].weight=0.5f;
  }
  for(i=25;i<33;i++)
  {
	tentacle_skin[i].weight=0.6f;
  }
  for(i=33;i<41;i++)
  {
	tentacle_skin[i].weight=1.0f;
  }
  for(i=41;i<NUM_VERTICES;i++)
  {
	tentacle_skin[i].weight=1.0f;
  }

  for(i=0;i<41;i++)
  {
	  tentacle_skin[i].bone1=0;
	  tentacle_skin[i].bone2=1;
  }
  for(i=41;i<NUM_VERTICES;i++)
  {
	  tentacle_skin[i].bone1=-1;
	  tentacle_skin[i].bone2=-1;
  }

  // setup bones
  bones[0].rotx=0.0f;
  bones[0].roty=0.0f;
  bones[0].rotz=0.0f;
  bones[0].xoffset=0.0f;
  bones[0].yoffset=0.0f;
  bones[0].zoffset=0.0f;
  bones[0].xextens=0.0f;
  bones[0].yextens=1.5f;
  bones[0].zextens=0.0f;
  bones[0].xjointx=1.0f;  // local X axis
  bones[0].xjointy=0.0f;
  bones[0].xjointz=0.0f;
  bones[0].zjointx=0.0f;  // local Z axis
  bones[0].zjointy=0.0f;
  bones[0].zjointz=1.0f;
  bones[0].parent=-1;
  bones[0].mat=1;

  bones[1].rotx=0.0f;
  bones[1].roty=0.0f;
  bones[1].rotz=0.0f;
  bones[1].xoffset=0.0f;
  bones[1].yoffset=1.5f;
  bones[1].zoffset=0.0f;
  bones[1].xextens=0.0f;
  bones[1].yextens=1.5f;
  bones[1].zextens=0.0f;
  bones[1].xjointx=1.0f;  // local X axis
  bones[1].xjointy=0.0f;
  bones[1].xjointz=0.0f;
  bones[1].zjointx=0.0f;  // local Z axis
  bones[1].zjointy=0.0f;
  bones[1].zjointz=1.0f;
  bones[1].parent=0;
  bones[1].mat=2;
  if(calculate_normals(vertices,NUM_VERTICES,indices,88,normals))
  for(i=0;i<NUM_VERTICES;i++)
  {
     tentacle_skin[i].nx=normals[i][0];
     tentacle_skin[i].ny=normals[i][1];
     tentacle_skin[i].nz=normals[i][2];
//  this should be replaced with normals inside of skinmesh
  }
  else exit(EXIT_FAILURE);    // something went seriously wrong
}

void build_matrices()
// update the local transformation matrices for all the bones of the model
{
	GLfloat yaw[16],pitch[16],roll[16],movforw[16],movback[16],transform[16],transform2[16];
	int i;
	for(i=0;i<NUM_BONES;i++)
	{
#pragma warning( disable : 4244 ) 
		roll[0]=COS(bones[i].rotz);
		roll[1]=SIN(bones[i].rotz);
		roll[4]=(-1)*SIN(bones[i].rotz);
		roll[5]=COS(bones[i].rotz);
		pitch[5]=COS(bones[i].rotx);
		pitch[6]=SIN(bones[i].rotx);
		pitch[9]=(-1)*SIN(bones[i].rotx);
		pitch[10]=COS(bones[i].rotx);
		yaw[0]=COS(bones[i].roty);
		yaw[2]=(-1)*SIN(bones[i].roty);
		yaw[8]=SIN(bones[i].roty);
		yaw[10]=COS(bones[i].roty);
#pragma warning( default : 4244 ) 
		roll[2]=0.0f;
		roll[3]=0.0f;
		roll[6]=0.0f;
		roll[7]=0.0f;
		roll[8]=0.0f;
		roll[9]=0.0f;
		roll[10]=1.0f;
		roll[11]=0.0f;
		roll[12]=0.0f;
		roll[13]=0.0f;
		roll[14]=0.0f;
		roll[15]=1.0f;
		pitch[0]=1.0f;
		pitch[1]=0.0f;
		pitch[2]=0.0f;
		pitch[3]=0.0f;
		pitch[4]=0.0f;
		pitch[7]=0.0f;
		pitch[8]=0.0f;
		pitch[11]=0.0f;
		pitch[12]=0.0f;
		pitch[13]=0.0f;
		pitch[14]=0.0f;
		pitch[15]=1.0f;
		yaw[1]=0.0f;
		yaw[3]=0.0f;
		yaw[4]=0.0f;
		yaw[5]=1.0f;
		yaw[6]=0.0f;
		yaw[7]=0.0f;
		yaw[9]=0.0f;
		yaw[11]=0.0f;
		yaw[12]=0.0f;
		yaw[13]=0.0f;
		yaw[14]=0.0f;
		yaw[15]=1.0f;
		movforw[0]=1.0f;
		movforw[1]=0.0f;
		movforw[2]=0.0f;
		movforw[3]=0.0f;
		movforw[4]=0.0f;
		movforw[5]=1.0f;
		movforw[6]=0.0f;
		movforw[7]=0.0f;
		movforw[8]=0.0f;
		movforw[9]=0.0f;
		movforw[10]=1.0f;
		movforw[11]=0.0f;
		movforw[12]=bones[i].xoffset;
		movforw[13]=bones[i].yoffset;
		movforw[14]=bones[i].zoffset;
		movforw[15]=1.0f;
		movback[0]=1.0f;
		movback[1]=0.0f;
		movback[2]=0.0f;
		movback[3]=0.0f;
		movback[4]=0.0f;
		movback[5]=1.0f;
		movback[6]=0.0f;
		movback[7]=0.0f;
		movback[8]=0.0f;
		movback[9]=0.0f;
		movback[10]=1.0f;
		movback[11]=0.0f;
		movback[12]=(-1)*bones[i].xoffset;
		movback[13]=(-1)*bones[i].yoffset;
		movback[14]=(-1)*bones[i].zoffset;
		movback[15]=1.0f;
		matxmat(transform,movforw,yaw);
		matxmat(transform2,transform,pitch);
		matxmat(transform,transform2,roll);
		matxmat(bones[i].tmatrix,transform,movback);
	}
}

void chain_up(int b,GLfloat *v,GLfloat *m)
// IK chain retrieval
// returns the coordinate that marks the end of the last bone
// in the IK chain (closest to the IK target), the end-effector
{
	GLfloat v1[4],v2[4],transform[16];
	int i;
	if(b==NUM_BONES-1)        // start of IK chain
	{
		chain_up(bones[b].parent,v,m);
		matxmat(transform,m,bones[b].tmatrix);
		v1[0]=bones[b].xextens+v[0];
		v1[1]=bones[b].yextens+v[1];
		v1[2]=bones[b].zextens+v[2];
		v1[3]=1.0f;
		matxvec(v2,transform,v1);
		v[0]=v2[0];
		v[1]=v2[1];
		v[2]=v2[2];
	}
	else if(bones[b].parent>-1)  // middle of IK chain
	{
		chain_up(bones[b].parent,v,m);
		matxmat(transform,m,bones[b].tmatrix);
		for(i=0;i<16;i++)
			m[i]=transform[i];
		v[0]+=bones[b].xextens;
		v[1]+=bones[b].yextens;
		v[2]+=bones[b].zextens;
	}
	else                    // end of IK chain
	{
		for(i=0;i<16;i++)
			m[i]=bones[b].tmatrix[i];
		v[0]+=bones[b].xextens;
		v[1]+=bones[b].yextens;
		v[2]+=bones[b].zextens;
	}
}

void skin()
// apply the mesh to the bones vertex by vertex
{
  GLfloat tempx1,tempx2,tempx3,tempx4,tempy1,tempy2;
  GLfloat tempy3,tempy4,tempz1,tempz2,tempz3,tempz4;
  GLfloat tvec[4],vec[4],transform[16];
  int i;   

  build_matrices();      // update local transformation matrices

  if(invkin)  // if IK is swiched on, then calculate it
  {
	  GLfloat a[3],b[3],c[4],pull[3],handle[3],v[4],w[4];
	  i=NUM_BONES-1;   // end of IK chain
      do
	  {
		  v[3]=1.0f;
		  handle[0]=0.0f;
		  handle[1]=0.0f;
		  handle[2]=0.0f;
		  chain_up(NUM_BONES-1,handle,transform);
		  // retrieve the end-effector of the IK chain

		  pull[0]=IKtarget[0]-handle[0];
		  pull[1]=IKtarget[1]-handle[1];
		  pull[2]=IKtarget[2]-handle[2];
		  // calculate the pull (force) vector
		  v[0]=bones[i].xextens;
		  v[1]=bones[i].yextens;
		  v[2]=bones[i].zextens;
		  matxvec(w,bones[i].tmatrix,v);
		  b[0]=w[0];
		  b[1]=w[1];
		  b[2]=w[2];
		  // find the current orientation of the bone in world space

		  v[0]=bones[i].xjointx;
		  v[1]=bones[i].xjointy;
		  v[2]=bones[i].xjointz;
		  matxvec(w,bones[i].tmatrix,v);
		  a[0]=w[0];
		  a[1]=w[1];
		  a[2]=w[2];
		  // find the current orientation of the X rotation axis in world space

		  v[0]=bones[i].zjointx;
		  v[1]=bones[i].zjointy;
		  v[2]=bones[i].zjointz;
		  matxvec(w,bones[i].tmatrix,v);
		  c[0]=w[0];
		  c[1]=w[1];
		  c[2]=w[2];
		  // find the current orientation of the Z rotation axis in world space
#pragma warning( disable : 4244 )
		  if(magnitude(pull)>0&&magnitude(pull)<3)	// when do we start/end calculating?
		  {
			  if(sinEnclosed(handle,pull)>0.02)     // termination condition
			  {
				  bones[i].rotx+=solvejoint(a,b,pull);  // solve IK for X
				  bones[i].rotz+=solvejoint(c,b,pull);  // solve IK for Z
			  }
			  
		  }
#pragma warning( default : 4244 )
		  i=bones[i].parent;      // move up to the next higher bone in the IK chain
		  build_matrices();
		  // in addition to the parent one should look for the start
		  // of the IK chain as loop terminator - todo once object loading is done
	  }while(i>(-1)); // as long as we haven't reached the end of the IK chain
  }

  for(i=0;i<NUM_VERTICES;i++)   // calculate skin vertex positions
  {
	  int bone1,bone2;

	  bone1=tentacle_skin[i].bone1;
	  bone2=tentacle_skin[i].bone2;

      vec[0]=vertices[i][0];
	  vec[1]=vertices[i][1];
	  vec[2]=vertices[i][2];
	  vec[3]=1.0f;           // homogeneous coordinate
	  if(bone1>(-1))
	  {
		  matxvec(tvec,bones[bone1].tmatrix,vec);
		  tempx1=tvec[0];
		  tempy1=tvec[1];
		  tempz1=tvec[2];
		  vec[0]=normals[i][0];
		  vec[1]=normals[i][1];
		  vec[2]=normals[i][2];
		  tempx3=tvec[0];
		  tempy3=tvec[1];
		  tempz3=tvec[2];
		  vec[0]=vertices[i][0];
		  vec[1]=vertices[i][1];
		  vec[2]=vertices[i][2];
		  if(bone2>(-1))
		  {
			  matxmat(transform,bones[bone1].tmatrix,bones[bone2].tmatrix);
			  matxvec(tvec,transform,vec);

			  tempx2=tvec[0];
			  tempy2=tvec[1];
			  tempz2=tvec[2];
			  vec[0]=normals[i][0];
			  vec[1]=normals[i][1];
			  vec[2]=normals[i][2];
			  matxvec(tvec,transform,vec);
			  
			  tempx4=tvec[0];
			  tempy4=tvec[1];
			  tempz4=tvec[2];
		  }
		  else
		  {
			  tempx2=vertices[i][0];
			  tempy2=vertices[i][1];
			  tempz2=vertices[i][2];
			  tempx4=normals[i][0];
			  tempy4=normals[i][1];
			  tempz4=normals[i][2];
		  }
	  }
	  else
	  {
		  tempx1=vertices[i][0];
		  tempy1=vertices[i][1];
		  tempz1=vertices[i][2];
		  tempx3=normals[i][0];
		  tempy3=normals[i][1];
		  tempz3=normals[i][2];
		  if(bone2==(-1))
		  {
			  tempx2=vertices[i][0];
			  tempy2=vertices[i][1];
			  tempz2=vertices[i][2];
			  tempx4=normals[i][0];
			  tempy4=normals[i][1];
			  tempz4=normals[i][2];
		  }
	  }
	  tentacle_skin[i].x=(tempx1*tentacle_skin[i].weight)+(tempx2*(1-tentacle_skin[i].weight));
	  tentacle_skin[i].y=(tempy1*tentacle_skin[i].weight)+(tempy2*(1-tentacle_skin[i].weight));
	  tentacle_skin[i].z=(tempz1*tentacle_skin[i].weight)+(tempz2*(1-tentacle_skin[i].weight));

	  tentacle_skin[i].nx=(tempx3*tentacle_skin[i].weight)+(tempx4*(1-tentacle_skin[i].weight));
	  tentacle_skin[i].ny=(tempy3*tentacle_skin[i].weight)+(tempy4*(1-tentacle_skin[i].weight));
	  tentacle_skin[i].nz=(tempz3*tentacle_skin[i].weight)+(tempz4*(1-tentacle_skin[i].weight));

  }
}


LONG WINAPI WndProc (HWND, UINT, WPARAM, LPARAM);
GLsizei glnWidth, glnHeight;
GLdouble gldAspect;

HPALETTE hPalette = NULL;
GLfloat nSize = 0.0f;
GLfloat nCol = 0.0f;
GLfloat nTop = 1.0f;
GLfloat nBottom = 0.0f;
int bFlag = 1;
char wndname[256]="";

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
    static char szAppName[] = "Mesh Skinning - Bendy Tentacle Demo";
    WNDCLASS wc;
    HWND hwnd;
    MSG msg;


    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szAppName;

    RegisterClass (&wc);

	// make sure that the window is not larger than the screen dimensions
	if(window_width>GetSystemMetrics(SM_CXSCREEN))
		window_width=GetSystemMetrics(SM_CXSCREEN);
	if(window_height>GetSystemMetrics(SM_CYSCREEN))
		window_height=GetSystemMetrics(SM_CYSCREEN);

    hwnd = CreateWindow (szAppName, szAppName,
        WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
        CW_USEDEFAULT, CW_USEDEFAULT, window_width, window_height,
        HWND_DESKTOP, NULL, hInstance, NULL);

    ShowWindow (hwnd, nCmdShow);
    UpdateWindow (hwnd);
    setup();

    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }
    return msg.wParam;
}

LONG WINAPI WndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static HDC hdc;
    static HGLRC hrc;
    PAINTSTRUCT ps;
    static double rotAngle = 0.0f;
    static double rotAngle2 = 0;
    static UINT nTimer;
	static int motion=1,motion2=1,motion3=1;
                
    switch (msg) {
    
    case WM_CREATE:
        // create rendering context and animation timer
        hdc = GetDC (hwnd);
        SetDCPixelFormat (hwnd,hdc);
        hrc = wglCreateContext (hdc);
        wglMakeCurrent (hdc, hrc);
        InitializeRC ();
        nTimer = SetTimer (hwnd, 1, 1, NULL);
        return 0;

    case WM_SIZE:
        glnWidth = (GLsizei) LOWORD (lParam);
        glnHeight = (GLsizei) HIWORD (lParam);

        return 0;

    case WM_PAINT:{
        // Draw the scene.
        BeginPaint (hwnd, &ps);
        DrawScene (hdc, rotAngle);  // do the rendering
        EndPaint (hwnd, &ps);       // finished

        }return 0;
#pragma warning( disable : 4244 )
	// standard Windows keyboard handling
	case WM_KEYUP:
		switch (wParam) {
		case    VK_LEFT:
		case   VK_RIGHT:
			motion=0;
			break;
		case      VK_UP:
		case    VK_DOWN:
			motion2=0;
			break;
		case   VK_PRIOR:
		case    VK_NEXT:
			motion3=0;
			break;
		}return 0;
	case WM_KEYDOWN:
		switch (wParam) {
		    case      'B':
				if(bones_off)
                   bones_off=0;
				else
                   bones_off=1;
				break;
			case      'F':
				if(fwkin)
				{
					fwkin=0;
					action=1;
					turntable=1;
					rotAngle=0;
					motion=speed1;
					motion2=speed1;
					motion3=speed1;
				}
				else
				{
					fwkin=1;
					invkin=0;
					action=0;
					turntable=0;
					rotAngle=90;
					motion=0;
					motion2=0;
					motion3=0;
				}
				break;
			case      'I':
				if(invkin)
				{
					invkin=0;
					action=1;
					turntable=1;
					rotAngle=0;
					motion=speed1;
					motion2=speed1;
					motion3=speed1;
				}
				else
				{
					fwkin=0;
					invkin=1;
					action=0;
					turntable=0;
					rotAngle=90;
					motion=0;
					motion2=0;
					motion3=0;
					IKtarget[0]=0.0f;
					IKtarget[1]=3.0f;
					IKtarget[2]=0.0f;
					bones[0].rotx=0.0f;
					bones[0].roty=0.0f;
					bones[0].rotz=0.0f;
					bones[1].rotx=0.0f;
					bones[1].roty=0.0f;
					bones[1].rotz=0.0f;
				}
				break;
		    case      'R':
				if(turntable)
                   turntable=0;
				else
                   turntable=1;
				break;
		    case      'P':
				if(action)
                   action=0;
				else
				{
                   action=1;
				   invkin=0;
				   fwkin=0;
				   motion=speed1;
				   motion2=speed1;
				   motion3=speed1;
				}
				break;
		    case      'N':
				if(normal_visibility)
                   normal_visibility=0;
				else
                   normal_visibility=1;
				break;
		    case      'O':
				if(optperf)
                   optperf=0;
				else
                   optperf=1;
				break;
		    case      'T':
				if(bone_visibility)
				{
                   bone_visibility=0;
				   bones_off=1;
				}
				else
				{
                   bone_visibility=1;
				   bones_off=0;
				}
				break;
#pragma warning( disable : 4305 )
			case VK_LEFT:
				if(fwkin)
					motion=-1*speed1;
				if(invkin)
					IKtarget[2]-=speed2;
				if(IKtarget[2]<(-2.5))   // clamp IK target
					IKtarget[2]=(-2.5);
				break;
			case VK_RIGHT:
				if(fwkin)
					motion=speed1;
				if(invkin)
					IKtarget[2]+=speed2;
				if(IKtarget[2]>2.5)   // clamp IK target
					IKtarget[2]=2.5;
				break;
			case VK_DOWN:
				if(fwkin)
					motion2=-1*speed1;
				if(invkin)
					IKtarget[1]-=speed2;
				if(IKtarget[1]<1.0)   // clamp IK target
					IKtarget[1]=1.0;
				break;
			case VK_UP:
				if(fwkin)
					motion2=speed1;
				if(invkin)
					IKtarget[1]+=speed2;
				if(IKtarget[1]>3.5)   // clamp IK target
					IKtarget[1]=3.5;
				break;
			case VK_PRIOR:
				if(fwkin)
					motion3=-1*speed1;
				if(invkin)
					IKtarget[0]-=speed2;
				if(IKtarget[0]<(-2.5))   // clamp IK target
					IKtarget[0]=(-2.5);
				break;
			case VK_NEXT:
				if(fwkin)
					motion3=speed1;
				if(invkin)
					IKtarget[0]+=speed2;
				if(IKtarget[0]>2.5)   // clamp IK target
					IKtarget[0]=2.5;
				break;
			case VK_ADD:
					speed1+=0.5;
					if(speed1>4)
						speed1=4.0f;
					speed2+=0.05;
					if(speed2>0.25)
						speed2=0.25f;
				break;
			case VK_SUBTRACT:
					speed1-=0.5f;
					if(speed1<0.5)
						speed1=0.5f;
					speed2+=0.05;
					if(speed2<0.01)
						speed2=0.01f;
				break;
#pragma warning( default : 4305 )
			case VK_ESCAPE:

				// exit the program on escape

				DestroyWindow(hwnd);
				break;
		}
		return 0;

    case WM_TIMER:
        // update animation and force a repaint
		if(turntable)
          rotAngle += speed1;
		
        if (rotAngle >= 360)
            rotAngle -= 360;

        if(fwkin)
		{
			bones[0].rotx+=motion;
			bones[1].rotx+=motion2;
			bones[1].rotz+=motion3;
		}

		if(action)
		{
		if (bones[0].rotx >= 17)
            motion = -1*(speed1+0.5);
		if (bones[0].rotx <= -17)
            motion = speed1+0.5;
		bones[0].rotx += motion;

		if (bones[1].rotx >= 95)
            motion2 = -1*(speed1+0.5);
		if (bones[1].rotx <= -35)
            motion2 = speed1+0.5;
		bones[1].rotx += motion2;

		if (bones[1].rotz >= 10)
            motion3 = -1*(speed1+0.5);
		if (bones[1].rotz <= -15)
            motion3 = speed1+0.5;
		bones[1].rotz += motion3;
		}
#pragma warning( default : 4244 )
		InvalidateRect (hwnd, NULL, FALSE);

        return 0;
    case WM_DESTROY:
        // end program
        wglMakeCurrent (NULL, NULL);
        wglDeleteContext (hrc);
        ReleaseDC (hwnd, hdc);
        if (hPalette != NULL)
            DeleteObject (hPalette);
        KillTimer (hwnd, nTimer);
        PostQuitMessage (0);
        return 0;
    }
    return DefWindowProc (hwnd, msg, wParam, lParam);
}

void SetDCPixelFormat (HWND hwnd,HDC hdc)
// function taken from an OpenGL program of unknown origin
{
    HANDLE hHeap;
    int nColors, i;
    LPLOGPALETTE lpPalette;
    BYTE byRedMask, byGreenMask, byBlueMask;

    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof (PIXELFORMATDESCRIPTOR),             // Size of this structure
        1,                                          // Version number
        PFD_DRAW_TO_WINDOW |                        // Flags
        PFD_SUPPORT_OPENGL |
        PFD_GENERIC_ACCELERATED|
        PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA,                              // RGBA pixel values
        24,                                         // 24-bit color
        0, 0, 0, 0, 0, 0,                           // Don't care about these
        0, 0,                                       // No alpha buffer
        0, 0, 0, 0, 0,                              // No accumulation buffer
        32,                                         // 32-bit depth buffer
        0,                                          // No stencil buffer
        0,                                          // No auxiliary buffers
        PFD_MAIN_PLANE,                             // Layer type
        0,                                          // Reserved (must be 0)
        0, 0, 0                                     // No layer masks
    };

    int nPixelFormat;

    nPixelFormat = ChoosePixelFormat (hdc, &pfd);
    SetPixelFormat (hdc, nPixelFormat, &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE) {
        nColors = 1 << pfd.cColorBits;
        hHeap = GetProcessHeap ();

        (LPLOGPALETTE) lpPalette = (LPLOGPALETTE)HeapAlloc (hHeap, 0,
            sizeof (LOGPALETTE) + (nColors * sizeof (PALETTEENTRY)));

        lpPalette->palVersion = 0x300;
        lpPalette->palNumEntries = nColors;

        byRedMask = (1 << pfd.cRedBits) - 1;
        byGreenMask = (1 << pfd.cGreenBits) - 1;
        byBlueMask = (1 << pfd.cBlueBits) - 1;

        for (i=0; i<nColors; i++) {
            lpPalette->palPalEntry[i].peRed =
                (((i >> pfd.cRedShift) & byRedMask) * 255) / byRedMask;
            lpPalette->palPalEntry[i].peGreen =
                (((i >> pfd.cGreenShift) & byGreenMask) * 255) / byGreenMask;
            lpPalette->palPalEntry[i].peBlue =
                (((i >> pfd.cBlueShift) & byBlueMask) * 255) / byBlueMask;
            lpPalette->palPalEntry[i].peFlags = 0;
        }

        hPalette = CreatePalette (lpPalette);
        HeapFree (hHeap, 0, lpPalette);

        if (hPalette != NULL) {
            SelectPalette (hdc, hPalette, FALSE);
            RealizePalette (hdc);
        }
    }
}

void InitializeRC (void)
// initialise the current rendering context
{
    GLfloat glfLightAmbient[] = { 0.1f, 0.1f, 0.1f, 1.0f };
    GLfloat glfLightDiffuse[] = { 1.2f, 1.2f, 1.2f, 1.0f };
    GLfloat glfLightSpecular[] = { 0.9f, 0.9f, 0.9f, 1.0f };

    // add light
    glLightfv (GL_LIGHT0, GL_AMBIENT, glfLightAmbient);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, glfLightDiffuse);
    glLightfv (GL_LIGHT0, GL_SPECULAR, glfLightSpecular);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);


    glEnable (GL_DEPTH_TEST);    
	glEnable (GL_CULL_FACE);
	glCullFace(GL_BACK);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE ,1);
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel(GL_SMOOTH);


}
void makeMaterial(GLenum glmode,GLfloat *rgb,GLfloat alpha)
{
 GLfloat m[4];
 m[0]=rgb[0];
 m[1]=rgb[1];
 m[2]=rgb[2];
 m[3]=alpha;
 glMaterialfv (GL_FRONT_AND_BACK,glmode,m);
}

void setMaterial(int i)
{
  GLfloat alpha=materials[i].alpha;
  makeMaterial (GL_AMBIENT, materials[i].ambient,alpha);
  makeMaterial (GL_DIFFUSE, materials[i].diffuse,alpha);
  makeMaterial (GL_SPECULAR, materials[i].specular,alpha);
  makeMaterial (GL_EMISSION, materials[i].glow,alpha);
  glMaterialf (GL_FRONT_AND_BACK,GL_SHININESS,materials[i].shine);
}

void DrawScene (HDC hdc, double rotAngle)
{


// set projection transformation
        gldAspect = (GLdouble) glnWidth / (GLdouble) glnHeight;

        glMatrixMode (GL_PROJECTION);
        glLoadIdentity ();
        gluPerspective (60.0,           // Field-of-view angle
                        gldAspect,      // Aspect ratio of viewing volume
                        1.0,            // Distance to near clipping plane
                        30.0);          // Distance to far clipping plane
// set viewport
        glViewport (0, 0, glnWidth, glnHeight);

    // clear color and depth buffers
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

// set world transformation
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    glTranslatef (0.0f, 0.0f, -12.0f);					// move camera back 12
	glTranslatef (0.0f, -3.0f, 0.0f);					// move camera down 3
    glRotatef (15.0f, 1.0f, 0.0f, 0.0f);				// tilt camera by 15 degrees

    glRotatef ((GLfloat) rotAngle, 0.0f, 1.0f, 0.0f);
    glScalef (nSize+2.5f,nSize+2.5f,nSize+2.5f);  


//  render the floor
	int i;
	setMaterial(0);
	if(!optperf)    // only texture the floor if the "optimal performance" flag is not set
	{
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texture[1]);
	}
	glBegin (GL_TRIANGLES); 
	  for(i=0;i<sizeof(floor_indices)/sizeof(floor_indices[0]);i++)
		{
		for(int j=0;j<3;j++)
		 {
		 int vi=floor_indices[i][j];
		 int ni=floor_indices[i][j+3];
		 int ti=floor_indices[i][j+6];
		 glNormal3f (floor_normals[ni][0],floor_normals[ni][1],floor_normals[ni][2]);
		 glTexCoord2f(floor_texture[ti][0],floor_texture[ti][1]);
		 glVertex3f (floor_vertices[vi][0],floor_vertices[vi][1],floor_vertices[vi][2]);
        }
       }
    glEnd ();
	if(!optperf)   // and we only need to switch off texturing if we actually use it
		glDisable(GL_TEXTURE_2D);


//  render the bones
if(bone_visibility && (!bones_off))    // should the bones be rendered?
{ // apparently so...
//  bone_0
	glPushMatrix();
	glRotatef((GLfloat) bones[0].roty,0.0,1.0,0.0);  // yaw
	glRotatef((GLfloat) bones[0].rotx,1.0,0.0,0.0);  // pitch
	glRotatef((GLfloat) bones[0].rotz,0.0,0.0,1.0);  // roll
	glBegin (GL_TRIANGLES);
	  setMaterial(bones[0].mat);
	  for(int i=0;i<sizeof(bone_indices)/sizeof(bone_indices[0]);i++)
		{
		for(int j=0;j<3;j++)
		 {
		 int vi=bone_indices[i][j];
		 int ni=bone_indices[i][j+3];
		 int ti=bone_indices[i][j+6];
		 glNormal3f (bone_normals[ni][0],bone_normals[ni][1],bone_normals[ni][2]);
		 glTexCoord2f(bone_textures[ti][0],bone_textures[ti][1]);
		 glVertex3f (bone_vertices[vi][0],bone_vertices[vi][1],bone_vertices[vi][2]);
        }
       }
    glEnd ();

//  bone_1
	glPushMatrix();
	glTranslatef(0.0,1.5,0.0);   // move into place
	glPushMatrix();
	glRotatef((GLfloat) bones[1].roty,0.0,1.0,0.0);  // yaw
	glRotatef((GLfloat) bones[1].rotx,1.0,0.0,0.0);  // pitch
	glRotatef((GLfloat) bones[1].rotz,0.0,0.0,1.0);  // roll
	glBegin (GL_TRIANGLES);
	  setMaterial(bones[1].mat);
	  for(i=0;i<sizeof(bone_indices)/sizeof(bone_indices[0]);i++)
	  {
		for(int j=0;j<3;j++)
		 {
		 int vi=bone_indices[i][j];
		 int ni=bone_indices[i][j+3];
		 int ti=bone_indices[i][j+6];
		 glNormal3f (bone_normals[ni][0],bone_normals[ni][1],bone_normals[ni][2]);
		 glTexCoord2f(bone_textures[ti][0],bone_textures[ti][1]);
		 glVertex3f (bone_vertices[vi][0],bone_vertices[vi][1],bone_vertices[vi][2]);
        }
       
       }
    glEnd ();
	glPopMatrix();
	glPopMatrix();
	glPopMatrix();
} // done with the bones

skin();

//  now render the tentacle
	  
for(i=0;i<88;i++)
{
	if(!bone_visibility)
	{
		if(!optperf)     // texturing is on
		{
			setMaterial(3);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, texture[0]);
		}
		else
			setMaterial(5);
		glBegin (GL_TRIANGLES);  // shaded mode
	}
	else
	{
		setMaterial(5);
		glBegin (GL_LINE_LOOP);  // wireframe mode
	}

	glNormal3f(tentacle_skin[tentacle_indices[i][2]].nx,tentacle_skin[tentacle_indices[i][2]].ny,tentacle_skin[tentacle_indices[i][2]].nz);
	glTexCoord2f(tentacle_skin[tentacle_indices[i][2]].u,tentacle_skin[tentacle_indices[i][2]].v);
	glVertex3f(tentacle_skin[tentacle_indices[i][2]].x,tentacle_skin[tentacle_indices[i][2]].y,tentacle_skin[tentacle_indices[i][2]].z);

	glNormal3f(tentacle_skin[tentacle_indices[i][1]].nx,tentacle_skin[tentacle_indices[i][1]].ny,tentacle_skin[tentacle_indices[i][1]].nz);
	glTexCoord2f(tentacle_skin[tentacle_indices[i][1]].u,tentacle_skin[tentacle_indices[i][1]].v);
    glVertex3f(tentacle_skin[tentacle_indices[i][1]].x,tentacle_skin[tentacle_indices[i][1]].y,tentacle_skin[tentacle_indices[i][1]].z);


	glNormal3f(tentacle_skin[tentacle_indices[i][0]].nx,tentacle_skin[tentacle_indices[i][0]].ny,tentacle_skin[tentacle_indices[i][0]].nz);
	glTexCoord2f(tentacle_skin[tentacle_indices[i][0]].u,tentacle_skin[tentacle_indices[i][0]].v);
    glVertex3f(tentacle_skin[tentacle_indices[i][0]].x,tentacle_skin[tentacle_indices[i][0]].y,tentacle_skin[tentacle_indices[i][0]].z);
    glEnd ();
	if(!optperf)
		glDisable(GL_TEXTURE_2D);
}

for(i=0;i<NUM_VERTICES;i++)
{
    setMaterial(4);
if(normal_visibility)
{
    // draw normals
    GLfloat xn,yn,zn;
	glPushMatrix();
	glBegin (GL_LINES);
    xn=tentacle_skin[i].x+(tentacle_skin[i].nx);
	yn=tentacle_skin[i].y+(tentacle_skin[i].ny);
	zn=tentacle_skin[i].z+(tentacle_skin[i].nz);
	glVertex3f(xn,yn,zn);
    glVertex3f(tentacle_skin[i].x,tentacle_skin[i].y,tentacle_skin[i].z);
	glEnd();
	glPopMatrix();
}
}
if(invkin)  // IK mode on?!
{
	GLfloat ikh[3],trans[16];
	ikh[0]=0.0f;
	ikh[1]=0.0f;
	ikh[2]=0.0f;
	setMaterial(3);
	glPushMatrix();
	glBegin(GL_LINES);           // draw IK target
#pragma warning( disable : 4244 )
    glVertex3f(IKtarget[0],IKtarget[1]+0.01,IKtarget[2]);
	glVertex3f(IKtarget[0],IKtarget[1]+0.05,IKtarget[2]);
#pragma warning( default : 4244 )
	glEnd();
	glPopMatrix();
	setMaterial(6);
	chain_up(1,ikh,trans);
	glPushMatrix();
	glBegin(GL_LINES);           // draw IK pull-vector
#pragma warning( disable : 4244 )
    glVertex3f(ikh[0],ikh[1],ikh[2]);
	glVertex3f(IKtarget[0],IKtarget[1],IKtarget[2]);
#pragma warning( default : 4244 )
	glEnd();
	glPopMatrix();
}
    glFlush();  // force OpenGL commands to finish
    SwapBuffers (hdc);      // swap screen buffers
}